//
// Copyright (c) 2004-2013 StreamBase Systems, Inc. All rights reserved.
//

package com.streambase.sb.contrib.liveview.test;


import java.text.MessageFormat;
import java.util.Date;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;

import org.slf4j.Logger;

import com.streambase.liveview.client.AbstractQueryListener;
import com.streambase.liveview.client.BeginDeleteEvent;
import com.streambase.liveview.client.BeginSnapshotEvent;
import com.streambase.liveview.client.EndDeleteEvent;
import com.streambase.liveview.client.EndSnapshotEvent;
import com.streambase.liveview.client.LiveViewConnection;
import com.streambase.liveview.client.LiveViewConnectionFactory;
import com.streambase.liveview.client.LiveViewException;
import com.streambase.liveview.client.LiveViewQueryType;
import com.streambase.liveview.client.Query;
import com.streambase.liveview.client.QueryClosedEvent;
import com.streambase.liveview.client.QueryConfig;
import com.streambase.liveview.client.QueryExceptionEvent;
import com.streambase.liveview.client.QueryListener;
import com.streambase.liveview.client.TupleAddedEvent;
import com.streambase.liveview.client.TupleRemovedEvent;
import com.streambase.liveview.client.TupleUpdatedEvent;
import com.streambase.liveview.client.internal.LiveViewConstants;
import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.TupleJSONUtil;
import com.streambase.sb.adapter.InputAdapter;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.PortCounts;
import com.streambase.sb.operator.TypecheckException;
import com.streambase.sb.util.Msg;
import com.streambase.sb.util.Util;

/**
 * Embedded adapter for testing query performance against LiveView.
 * The LiveViewQuery adapter should be used for query tables and validating/consuming the data. 
 *
 * Input schema:
 *   tableName : string
 *   queryID : string
 *   query : string
 *   	empty string is get everything
 *   Unregister : boolean
 *   	true means remove query. all other fields other than queryID are ignored.
 *   snapshotOnly : boolean
 *   	null or false mean a continuous query
 */
public class LiveViewQueryTest extends InputAdapter implements Parameterizable, Runnable {

    public static final int QUERY_OUT_PORT = 1;
    public static final int QUERY_IN_PORT = 0;

	protected Logger logger;
	
    public static final int STATUS_OUT_PORT = 0;
    
	// Status field names
    public static final String TYPE_FIELD = "Type";
    public static final String OBJECT_FIELD = "Object";
    public static final String ACTION_FIELD = "Action";
    public static final String MESSAGE_FIELD = "Message";
    public static final String TUPLE_FIELD = "Tuple";
    public static final String TIME_FIELD = "Time";
    
    // status types
    public static final String CONNECTION_STATUS_TYPE="CONNECTION";
    public static final String ERROR_STATUS_TYPE="ERROR";
    public static final String QUERY_STATUS_TYPE="QUERY";
    public static final String PUBLISH_STATUS_TYPE="PUBLISH";
	public static final String ADMIN_TYPE = "ADMIN";
    
    // status actions
    public static final String CONNECTED_ACTION="CONNECTED";
    public static final String FAILED_ACTION="FAILED";
    public static final String ERROR_ACTION="ERROR";
    public static final String REGISTER_QUERY_ACTION="REGISTER_QUERY";
    public static final String UNREGISTER_QUERY_ACTION="UNREGISTER_QUERY";
    public static final String SNAPSHOT_BEGIN_ACTION="SNAPSHOT_BEGIN";
    public static final String SNAPSHOT_END_ACTION="SNAPSHOT_END";
    public static final String DELETE_BEGIN_ACTION="DELETE_BEGIN";
    public static final String DELETE_END_ACTION="DELETE_END";
    
    public final static String CLIENT_INFO_TAG="?" + LiveViewConstants.CLIENT_INFO;  
    
    protected Schema statusSchema=null;
    
    protected LiveViewConnection lvServer=null;
    private String lvURI=null;
    protected AtomicBoolean connected = new AtomicBoolean(false);
    
    private String displayName = "LiveView Query Tester";

    private Map<String, Query> activeQueries = new ConcurrentHashMap<String, Query>();
    private AtomicBoolean done = new AtomicBoolean(false);
    private static final int RECONNECT_SLEEP=Integer.parseInt((Util.getSystemProperty("streambase.adapter.test.lvquery.reconnectms", "1000")));

    private Schema querySchema;

    // Field names
    public static final String SELECT = "Select";
    public static final String QUERY = "Query";
    public static final String SNAPSHOT = "Snapshot";
    public static final String SNAPSHOT_ONLY = "SnapshotOnly";
    public static final String UNREGISTER_QUERY = "Unregister";

    public static final String TABLENAME_FIELD = "TableName";
    public static final String QUERYID_FIELD = "QueryName";
    public static final String CQS_INTERNALID_FIELD = "CQSInternalID";
    public static final String CQS_REFERRED_COUNT_FIELD = "CQSReferredCount";
    public static final String CQS_SCOPE_TRANSITION_FIELD = "CQSScopeTransition";
    public static final String RESULT_DATA_FIELD = "Results";
    
    // Scope definitions
    public static final int SCOPE_ADD=0;
    public static final int SCOPE_DELETE=1;
    public static final int SCOPE_UPDATE=2;
   
    // Schema.Field members use to set tuple fields
    private Schema.Field tableNameField;
    private Schema.Field queryIDField;
    private Schema.Field cqsInternalIDField;
    private Schema.Field cqsScopeField;
    private Schema.Field resultsField;
    private Schema.Field querySnapshotField;

    private Schema.Field queryInTableNameField;
    private Schema.Field queryInQueryIDField;
    private Schema.Field queryInSelectField;
    private Schema.Field queryInQueryField;
    private Schema.Field queryInSnapshotOnlyField;
    private Schema.Field queryInDeleteQueryField;
    
    // The query listener used by all queries
    private class MyQueryListener extends AbstractQueryListener {
        private String queryID;
        private String table;
        private String uniqueID;
        private Tuple resultTuple;
        private boolean inSnapshot = false;

        MyQueryListener(String table, String queryID) {
        	this.table = table;
            this.queryID = queryID;
            uniqueID = getUniqueID(table, queryID);
            resultTuple = querySchema.createTuple();
        }
        
        void sendResult(long CQSInternalID, int scope, Tuple newTuple, Tuple oldTuple) {
        	if (!done.get()) {
        		try {
        			resultTuple.clear();
        			resultTuple.setString(tableNameField, table);
	                resultTuple.setString(queryIDField, queryID);
	            	resultTuple.setLong(cqsInternalIDField, CQSInternalID);
	            	resultTuple.setInt(cqsScopeField, scope);
	            	// just serialize the type into a JSON string
	            	if (newTuple != null) {
	            		resultTuple.setString(resultsField, TupleJSONUtil.toJSONListString(newTuple));
	            	}
	            	
	                resultTuple.setBoolean(querySnapshotField, inSnapshot);
	                
	                sendOutputAsync(QUERY_OUT_PORT, resultTuple);
		        	
	        	} catch (Exception e) {
	        		warn(Msg.format("Error sending query: table {0}, name {1} {2}", table, queryID, e.getMessage()));
	        	}
        	}
        }

        @Override
        public void tupleAdded(final TupleAddedEvent event) {
            sendResult(event.getKey(), SCOPE_ADD, event.getTuple(), null);
        }

        @Override
        public void tupleRemoved(final TupleRemovedEvent event) {
            sendResult(event.getKey(), SCOPE_DELETE, null, event.getTuple());
        }

        @Override
        public void tupleUpdated(final TupleUpdatedEvent event) {
            sendResult(event.getKey(), SCOPE_UPDATE, event.getTuple(), event.getOld());
        }

        @Override
        public void snapshotBegin(BeginSnapshotEvent event) {
            super.snapshotBegin(event);
            inSnapshot = true;
            sendStatusTuple(QUERY_STATUS_TYPE, SNAPSHOT_BEGIN_ACTION, uniqueID);
        }

        @Override
        public void snapshotEnd(EndSnapshotEvent event) {
            super.snapshotEnd(event);
            inSnapshot = false;
            sendStatusTuple(QUERY_STATUS_TYPE, SNAPSHOT_END_ACTION, uniqueID);
        }
        
    	@Override
    	public void exceptionRaised(final QueryExceptionEvent event) {
    		String msg=Msg.format("QueryName: {0} got exception: {1}", uniqueID, event.getException().getMessage());
    		sendStatusTuple(QUERY_STATUS_TYPE, ERROR_ACTION, msg);
    		warn(msg);
    		
    		unregisterQuery(table, queryID, false);
    		
    		// See if the connection is gone
    		isConnectionOK();
    	}

    	/**
    	 * {@inheritDoc}
    	 */
    	@Override
    	public void queryClosed(final QueryClosedEvent event) {
    		unregisterQuery(table, queryID, false);
    		logger.debug(Msg.format("Query: {0} closed", uniqueID));
    	}

		@Override
		public void deleteBegin(BeginDeleteEvent event) {
			logger.warn(Msg.format("Query: {0} unexpected BeginDeleteEvent", uniqueID));
		}

		@Override
		public void deleteEnd(EndDeleteEvent event) {
			logger.warn(Msg.format("Query: {0} unexpected EndDeleteEvent", uniqueID));			
		}
    }

    public LiveViewQueryTest() {
        super();
        logger = getLogger();
        setPortHints(1, 2);
        setDisplayName(displayName);
        setShortDisplayName(this.getClass().getSimpleName());
        setUri("lv://localhost");
    }

    
    public void typecheck() throws TypecheckException {
        requireInputPortCount(1);

        if (Util.isEmpty(lvURI)) {
        	throw new TypecheckException("Must specify URI, or encapulated LiveView");
        }
        
        // get the required input fields.
        try {
        	queryInTableNameField=getInputSchema(QUERY_IN_PORT).getField(TABLENAME_FIELD);
            queryInQueryIDField=getInputSchema(QUERY_IN_PORT).getField(QUERYID_FIELD);
            queryInSelectField=getInputSchema(QUERY_IN_PORT).getField(SELECT);
            queryInQueryField=getInputSchema(QUERY_IN_PORT).getField(QUERY);
            queryInSnapshotOnlyField=getInputSchema(QUERY_IN_PORT).getField(SNAPSHOT_ONLY);
            queryInDeleteQueryField=getInputSchema(QUERY_IN_PORT).getField(UNREGISTER_QUERY);
        } catch (TupleException e) {
        	throw new TypecheckException(e);
        }

        querySchema = new Schema(null,
        		Schema.createField(DataType.STRING, TABLENAME_FIELD),
                Schema.createField(DataType.STRING, QUERYID_FIELD),
                Schema.createField(DataType.LONG, CQS_INTERNALID_FIELD),
                Schema.createField(DataType.INT, CQS_SCOPE_TRANSITION_FIELD),
                Schema.createField(DataType.BOOL, SNAPSHOT),
                Schema.createField(DataType.STRING, RESULT_DATA_FIELD)
                );


        setStatusSchema(querySchema);
        setOutputSchema(STATUS_OUT_PORT, statusSchema);
        setOutputSchema(QUERY_OUT_PORT, querySchema);
    }

    @Override
    public PortCounts getPortCounts() throws TypecheckException {
        return new PortCounts(1, 2);
    }

    public void init() throws StreamBaseException {
        super.init();

        // Register the object so it will be run as a thread managed by StreamBase.
        registerRunnable(this, true);
        
    	tableNameField=querySchema.getField(TABLENAME_FIELD);
        queryIDField = querySchema.getField(QUERYID_FIELD);
        cqsInternalIDField = querySchema.getField(CQS_INTERNALID_FIELD);
        cqsScopeField = querySchema.getField(CQS_SCOPE_TRANSITION_FIELD);
        resultsField= querySchema.getField(RESULT_DATA_FIELD);
        querySnapshotField = querySchema.getField(SNAPSHOT);
    }

    public void shutdown() {
    	try {
	    	// Close all active queries
	    	for (Query q : activeQueries.values()) {
	    		q.close();
	    	}
    	} finally {
	    	connected.set(false);
			if (lvServer != null) {
				lvServer.close();
				lvServer = null;
			}
    	}
    }

    @Override
    public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
        super.processTuple(inputPort, tuple);

        if(inputPort == QUERY_IN_PORT) {
        	
        	if (!connected.get()) {
        		String msg=Msg.format("Not connected to {0}, tuple: {1}", lvURI, tuple);
            	warn(msg);
            	sendStatusTuple(CONNECTION_STATUS_TYPE, ERROR_STATUS_TYPE, msg);
            	return;
        	}
        	
            boolean unregister = false;
            boolean snapshotOnly = false;

            if(!tuple.isNull(queryInDeleteQueryField))
            	unregister = tuple.getBoolean(queryInDeleteQueryField);

            if(!tuple.isNull(queryInSnapshotOnlyField))
            	snapshotOnly = tuple.getBoolean(queryInSnapshotOnlyField);

            if(unregister && snapshotOnly)
            	warn(Msg.format("Since unregister is set, ignoring snapshotOnly and query fields: {0}", tuple));

            if (tuple.isNull(queryInTableNameField)) {
            	throw new StreamBaseException("TableName can not be null");
            }
            String table = tuple.getString(queryInTableNameField);
            
            if (tuple.isNull(queryInQueryIDField)) {
            	throw new StreamBaseException("QueryID can not be null");
            }
            String queryID = tuple.getString(queryInQueryIDField);

            if(!unregister) {
                String query="true";
                if (!tuple.isNull(queryInQueryField)) {
                	query = tuple.getString(queryInQueryField);
                }
                String select="*";
                if (!tuple.isNull(queryInSelectField)) {
                	select = tuple.getString(queryInSelectField);
                }
            	addQuery(select, query, queryID, table, snapshotOnly);
            } else {
            	unregisterQuery(table, queryID, false);
            }
        }
    }

	public void run() {
		boolean reportedDisconnect=false;

    	while(shouldRun()) {
    		if(!connected.get()) {
    			try {
    				connect();
    	    		reportedDisconnect=false;
    	    		connected.set(true);        	    	
    			} catch (Exception e1) {
    				String msg=Msg.format("Connection to {0} failed: {1}", lvURI, e1.getMessage());
    				logger.debug(MessageFormat.format("Error connecting to LiveView server {0}", lvURI), e1);
    				if (!reportedDisconnect) {
    					sendStatusTuple(CONNECTION_STATUS_TYPE, FAILED_ACTION, msg);
    					reportedDisconnect=true;
    				}
    			}
    		}

    		try {
    			Thread.sleep(RECONNECT_SLEEP);
            } catch (InterruptedException e) {
            	// this can happen on a normal shutdown
            }
        }
    	done.set(true);
    }

    private void addQuery(String select, String query, String queryID, String table, boolean snapshotOnly) throws StreamBaseException {

        String uniqueID = getUniqueID(table, queryID);
        
        if (activeQueries.containsKey(uniqueID)) {
        	String msg=MessageFormat.format("Query {0} is already in use, ignoring new query", uniqueID);
        	sendStatusTuple(QUERY_STATUS_TYPE, ERROR_ACTION, msg);
        	warn(msg);
        	return;
        }

        LiveViewQueryType qtype = LiveViewQueryType.SNAPSHOT_AND_CONTINUOUS;

        if(snapshotOnly)
        	qtype = LiveViewQueryType.SNAPSHOT;

        String sqlQuery = Msg.format("select {0} from {1} where {2}", select, table, query);
        debug(Msg.format("Adding Query id {0} : {1}", uniqueID, sqlQuery));

        QueryConfig queryConfig = new QueryConfig().setQueryString(sqlQuery).setQueryType(qtype);
        QueryListener queryListener=new MyQueryListener(table, queryID);
        Query regquery=lvServer.registerQuery(queryConfig, queryListener);

        sendStatusTuple(QUERY_STATUS_TYPE, REGISTER_QUERY_ACTION, getUniqueID(table, queryID));
        
        activeQueries.put(uniqueID, regquery);
    }
    
    private void unregisterQuery(String table, String queryID, boolean reportNotFound) {
        String uniqueID = getUniqueID(table, queryID);
        
        Query regquery = activeQueries.get(uniqueID);

        if(regquery == null) {
        	if (reportNotFound) {
	        	String msg=Msg.format("Cant unregister unknown query: {0}", uniqueID);
	        	sendStatusTuple(QUERY_STATUS_TYPE, ERROR_ACTION, msg);
	        	info(msg);
        	}
        	return;
        } else {
        	debug(Msg.format("Unregistering query {0}", uniqueID));
        }
        activeQueries.remove(uniqueID);
        sendStatusTuple(QUERY_STATUS_TYPE, UNREGISTER_QUERY_ACTION, uniqueID);
        regquery.close();
    }
    
    private String getUniqueID(String table, String queryID) {
    	return Msg.format("{0}-{1}", table, queryID);
    }
    
    public void closeConnection() {
        lvServer.close();
    	lvServer=null;
    	connected.set(false);
    }
    
    protected void setStatusSchema(Schema s) {
        Schema.Field[] STATUS_FIELDS = {
                new Schema.Field(TYPE_FIELD, CompleteDataType.forString()),
                new Schema.Field(OBJECT_FIELD, CompleteDataType.forString()),
                new Schema.Field(ACTION_FIELD, CompleteDataType.forString()),
                new Schema.Field(MESSAGE_FIELD, CompleteDataType.forString()),
                new Schema.Field(TIME_FIELD, CompleteDataType.forTimestamp()),
            };
            statusSchema=new Schema(null, STATUS_FIELDS);
    }
    
    protected void connect() throws StreamBaseException {
    	logger.debug(Msg.format("Connecting to {0}", lvURI));
    	lvServer = LiveViewConnectionFactory.getConnection(Msg.format("{0}{1}={2}", lvURI, CLIENT_INFO_TAG, getName()));
    	logger.debug(Msg.format("Connected to {0}", lvURI));
    	sendStatusTuple(CONNECTION_STATUS_TYPE, CONNECTED_ACTION, lvURI);
    }
    
    protected boolean isConnectionOK() {
    	boolean good=true;
    	Exception savedE=null;
    	try {
			if (lvServer.getVersion() == null) {
		    	good=false;
			}
			
		} catch (LiveViewException e) {
			savedE=e;
			logger.debug("isConnectionOK failed", e);
			good=false;
		}
    	
    	if (!good) {
    		String msg=Msg.format("Connection failed: {0} Exception: {1}", lvURI, (savedE==null) ? "null Version returned" : savedE.getMessage());
        	sendStatusTuple(CONNECTION_STATUS_TYPE, ERROR_ACTION, msg);
        	logger.warn(msg);

	    	connected.set(false);
	    	lvServer.close();
	    	lvServer=null;
    	}
    	return good;
    }
    
    protected void sendStatusTuple(String type, String action, String message) {
        
        try {
            Tuple t=statusSchema.createTuple();
            t.setString(TYPE_FIELD, type);
            t.setString(MESSAGE_FIELD, message);
            t.setString(OBJECT_FIELD, getName());
            t.setString(ACTION_FIELD, action);
            t.setTimestamp(TIME_FIELD, new Timestamp(new Date()));
            sendOutput(STATUS_OUT_PORT, t);
        } catch (TupleException e) {
            logger.error(Msg.format("{0}: Error creating status tuple", getName()), e);
        } catch (StreamBaseException e) {
            logger.error(Msg.format("{0}: Exception sending status tuple", getName()), e);
        }
    }
  
    protected void error(String msg) {
    	logger.error(Msg.format("{0}: {1}", getName(), msg));
    }
    
    protected void warn(String msg) {
    	logger.warn(Msg.format("{0}: {1}", getName(), msg));
    }
    
    protected void info(String msg) {
    	logger.info(Msg.format("{0}: {1}", getName(), msg));
    }
    
    protected void debug(String msg) {
    	logger.debug(Msg.format("{0}: {1}", getName(), msg));
    }

    /***************************************************************************************
     * The getter and setter methods provided by the Parameterizable object.               *
     * StreamBase Studio uses them to determine the name and type of each property         *
     * and obviously, to set and get the property values.                                  *
     ***************************************************************************************/

    public void setUri(String lvURI) {
    	this.lvURI = lvURI;
    }
    public String getUri() {
        return lvURI;
    }
    
}
